import paramiko
from typing import Optional
import abc

_SQUEUE_FORMAT = "%a|%A|%T|%B|%e"
_SQUEUE_FORMAT_KEYS = ("ACCOUNT", "JOBID", "STATE", "EXEC_HOST", "END_TIME")
_SACCT_FORMAT_KEYS = ("ACCOUNT", "JOBID", "STATE", "END_TIME")


class SlurmError(Exception):
    pass


class Session(abc.ABC):
    @abc.abstractmethod
    def exec_cmd(self, cmd: str):
        pass

    def sbatch(self, jobexec: str, **kwargs) -> int:
        args = " ".join(
            map(lambda kv: f"--{kv[0]}={kv[1]}" if kv[1] else f"--{kv[0]}", kwargs.items())
        )
        out, err, exit_code = self.exec_cmd(f"sbatch {jobexec} " + args)
        if exit_code != 0:
            raise SlurmError(err)
        return int(out.strip().split(" ")[-1])

    def squeue(self, user_id: Optional[str] = None, job_id: Optional[int] = None):
        uid_flag = ""
        job_id_flag = ""
        if user_id is not None:
            uid_flag = f"-u {user_id} "
        if job_id is not None:
            job_id_flag = f"-j {job_id} "

        args = f'squeue {uid_flag}{job_id_flag} -o "{_SQUEUE_FORMAT}" -h'
        out, err, exit_code = self.exec_cmd(args)
        if exit_code != 0:
            raise SlurmError(err)
        # parsing squeue output
        out = out.strip()
        if out == "":
            return []
        return [
            dict(zip(_SQUEUE_FORMAT_KEYS, line.split("|"))) for line in out.split("\n")
        ]

    def scancel(self, job_id: int):
        _, err, exit_code = self.exec_cmd(f"scancel {job_id}")
        if exit_code != 0:
            raise SlurmError(err)

    def sacct(self, job_id=None, allocations=True):
        # use -X flag (maybe)
        args = ["sacct", "-n", "-P", "--format", "User,JobID,State,End"]
        if allocations:
            args.append("-X")
        if job_id:
            args.append(f"-j {job_id}")

        out, err, return_code = self.exec_cmd(" ".join(args))
        if return_code != 0:
            raise SlurmError(err)
        out = out.strip()
        if out == "":
            return []
        entries = []
        for line in out.split("\n"):
            entry = dict(zip(_SACCT_FORMAT_KEYS, line.split("|")))
            entries.append(entry)
        return entries


class SSHSession(Session):
    def __init__(self, host, **kwargs):
        self.client = paramiko.SSHClient()
        self.client.load_system_host_keys()
        self.client.connect(host, **kwargs)

    def exec_cmd(self, cmd, timeout=15):
        _, stdout, stderr = self.client.exec_command(cmd, timeout=timeout)
        stdout_output = stdout.read().decode("utf8")
        stderr_output = stderr.read().decode("utf8")
        exit_code = stdout.channel.recv_exit_status()
        return stdout_output, stderr_output, exit_code


class LocalSession(Session):
    def __init__(self):
        raise NotImplementedError

    def exec_cmd(self, cmd):
        raise NotImplementedError



from dataclasses import dataclass
import datetime
import time
from .session import SlurmError


class JobNotFound(SlurmError):
    pass


def _squeue_entry_is_done(state):
    return state in {
        "BOOT_FAIL",
        "CANCELLED",
        "COMPLETED",
        "DEALINE",
        "FAILED",
        "NODE_FAIL",
        "OUT_OF_MEMORY",
        "PREEMPTED",
        "TIMEOUT",
    }


def _parse_date(datestr):
    if datestr in ("Unknown", "N/A"):
        return None
    else:
        return datetime.datetime.fromisoformat(datestr)


@dataclass
class Job:
    account: str
    job_id: str
    state: str
    end_time: datetime.datetime | None

    def _update(self, session, raise_not_found=True):
        queue_entry = session.squeue(job_id=self.job_id)
        if len(queue_entry) > 0:
            self.state = queue_entry[0]["STATE"]
            self.end_time = _parse_date(queue_entry[0]["END_TIME"])
            return
        acct_entry = session.sacct(job_id=self.job_id)
        if len(acct_entry) > 0:
            self.state = acct_entry[0]["STATE"]
            self.end_time = _parse_date(acct_entry[0]["END_TIME"])
            return
        if raise_not_found:
            raise JobNotFound()

    @classmethod
    def from_queue(cls, squeue_output):
        return cls(
            account=squeue_output["ACCOUNT"],
            job_id=squeue_output["JOBID"],
            state=squeue_output["STATE"],
            end_time=_parse_date(squeue_output["END_TIME"]),
        )

    def poll_to_completion(self, session, interval=10):
        while True:
            squeue_entries = session.squeue(job_id=self.job_id)
            if len(squeue_entries) == 0:
                self._update(session)
                return True
            time.sleep(interval)

    def cancel(self, session):
        session.scancel(job_id=self.job_id)

    @property
    def running(self):
        return self.state == "RUNNING"

    @property
    def pending(self):
        return self.state == "PENDING"

    @property
    def done(self):
        return _squeue_entry_is_done(self.state)



from .job import Job, JobNotFound
from .session import LocalSession, SSHSession


class SlurmAcct:
    def __init__(self, session):
        self.session = session

    def lookup(self, job):
        if type(job) is int:
            job_id = job
        else:
            job_id = job.job_id
        acct = self.session.sacct(job_id=job_id)
        if len(acct) > 0:
            return Job.from_queue(acct[0])
        else:
            return None

    def all(self):
        return [Job.from_queue(e) for e in self.session.sacct()]


class SlurmQueue:
    def __init__(self, session):
        self.session = session

    def submit(self, jobspec, wait=False, **kwargs):
        if type(jobspec) is str:
            jobexec = jobspec
            jobargs = {}
        else:
            jobexec = jobspec.exec
            jobargs = jobspec.construct_args()
        return self.session.sbatch(jobexec, **jobargs)

    def lookup(self, job):
        if type(job) is int:
            job_id = job
        else:
            job_id = job.job_id

        in_queue = self.session.squeue(job_id=job_id)
        if len(in_queue) > 0:
            return Job.from_queue(in_queue[0])
        else:
            return None

    def all(self):
        return [Job.from_queue(e) for e in self.session.squeue()]


# TODO: integrate user_id (required for squeue and sacct)
class Slurm:
    def __init__(
            self, session=None, host=None, username=None, password=None, local=False
    ):
        if session:
            self.session = session
        elif host:
            sshargs = {}
            if username:
                sshargs["username"] = username
            if password:
                sshargs["password"] = password
            self.session = SSHSession(host, **sshargs)
        elif local:
            self.session = LocalSession()
        else:
            raise ValueError(
                "Slurm.__init__ needs to be called with argument session or ssh_host set, or local set to True"
            )
        self.queue = SlurmQueue(self.session)
        self.acct = SlurmAcct(self.session)

    def submit(self, jobspec, wait=False, **kwargs):
        job_id = self.queue.submit(jobspec, wait=wait, **kwargs)
        submitted = self.lookup(job_id)
        if submitted is None:
            raise JobNotFound
        else:
            return submitted

    def queued(self):
        return self.queue.all()

    def lookup(self, job):
        in_queue = self.queue.lookup(job)
        if in_queue:
            return in_queue
        return self.acct.lookup(job)



from .job import Job, JobNotFound

from copy import deepcopy


class Jobspec:
    def __init__(self, jobexec):
        self.exec = jobexec
        self._exec_args = {}
        self._dependencies = {}

    def construct_args(self):
        args = deepcopy(self._exec_args)
        deps = []
        for kind, ids in self._dependencies:
            deps.append(f"{kind}:" + ":".join(ids))
        if deps:
            args["dependencies"] = ",".join(deps)
        return args

    def after_ok(self, job):
        if "afterok" in self._dependencies:
            self._dependencies["afterok"].append(job.id)
        else:
            self._dependencies["afterok"] = [job.id]
        return self

    def exclusive(self):
        self._exec_args["exlusive"] = ""
        return self

    def nodes(self, n):
        self._exec_args["nodes"] = n
        return self

    def ntasks(self, n):
        self._exec_args["ntasks"] = n
        return self

    def ntasks_per_node(self, n):
        self._exec_args["ntasks-per-node"] = n
        return self

    def cpus_per_task(self, n):
        self._exec_args["cpus-per-task"] = n
        return self

    def mem_per_cpu(self, n):
        self._exec_args["mem-per-cpu"] = n
        return self

    def job_name(self, name):
        self._exec_args["job-name"] = name
        return self

    def output(self, output):
        self._exec_args["output"] = output
        return self
